(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
	typeof define === 'function' && define.amd ? define(factory) :
	(global.VirtualList = factory());
}(this, (function () { 'use strict';

	function noop() {}

	function assign(tar, src) {
		for (var k in src) tar[k] = src[k];
		return tar;
	}

	function assignTrue(tar, src) {
		for (var k in src) tar[k] = 1;
		return tar;
	}

	function append(target, node) {
		target.appendChild(node);
	}

	function insert(target, node, anchor) {
		target.insertBefore(node, anchor);
	}

	function detachNode(node) {
		node.parentNode.removeChild(node);
	}

	function createElement(name) {
		return document.createElement(name);
	}

	function createText(data) {
		return document.createTextNode(data);
	}

	function addListener(node, event, handler) {
		node.addEventListener(event, handler, false);
	}

	function removeListener(node, event, handler) {
		node.removeEventListener(event, handler, false);
	}

	function setStyle(node, key, value) {
		node.style.setProperty(key, value);
	}

	function addResizeListener(element, fn) {
		if (getComputedStyle(element).position === 'static') {
			element.style.position = 'relative';
		}

		const object = document.createElement('object');
		object.setAttribute('style', 'display: block; position: absolute; top: 0; left: 0; height: 100%; width: 100%; overflow: hidden; pointer-events: none; z-index: -1;');
		object.type = 'text/html';

		let win;

		object.onload = () => {
			win = object.contentDocument.defaultView;
			win.addEventListener('resize', fn);
		};

		if (/Trident/.test(navigator.userAgent)) {
			element.appendChild(object);
			object.data = 'about:blank';
		} else {
			object.data = 'about:blank';
			element.appendChild(object);
		}

		return {
			cancel: () => {
				win && win.removeEventListener && win.removeEventListener('resize', fn);
				element.removeChild(object);
			}
		};
	}

	function destroyBlock(block, lookup) {
		block.d(1);
		lookup[block.key] = null;
	}

	function updateKeyedEach(old_blocks, component, changed, get_key, dynamic, ctx, list, lookup, node, destroy, create_each_block, intro_method, next, get_context) {
		var o = old_blocks.length;
		var n = list.length;

		var i = o;
		var old_indexes = {};
		while (i--) old_indexes[old_blocks[i].key] = i;

		var new_blocks = [];
		var new_lookup = {};
		var deltas = {};

		var i = n;
		while (i--) {
			var child_ctx = get_context(ctx, list, i);
			var key = get_key(child_ctx);
			var block = lookup[key];

			if (!block) {
				block = create_each_block(component, key, child_ctx);
				block.c();
			} else if (dynamic) {
				block.p(changed, child_ctx);
			}

			new_blocks[i] = new_lookup[key] = block;

			if (key in old_indexes) deltas[key] = Math.abs(i - old_indexes[key]);
		}

		var will_move = {};
		var did_move = {};

		function insert(block) {
			block[intro_method](node, next);
			lookup[block.key] = block;
			next = block.first;
			n--;
		}

		while (o && n) {
			var new_block = new_blocks[n - 1];
			var old_block = old_blocks[o - 1];
			var new_key = new_block.key;
			var old_key = old_block.key;

			if (new_block === old_block) {
				// do nothing
				next = new_block.first;
				o--;
				n--;
			}

			else if (!new_lookup[old_key]) {
				// remove old block
				destroy(old_block, lookup);
				o--;
			}

			else if (!lookup[new_key] || will_move[new_key]) {
				insert(new_block);
			}

			else if (did_move[old_key]) {
				o--;

			} else if (deltas[new_key] > deltas[old_key]) {
				did_move[new_key] = true;
				insert(new_block);

			} else {
				will_move[old_key] = true;
				o--;
			}
		}

		while (o--) {
			var old_block = old_blocks[o];
			if (!new_lookup[old_block.key]) destroy(old_block, lookup);
		}

		while (n) insert(new_blocks[n - 1]);

		return new_blocks;
	}

	function getSpreadUpdate(levels, updates) {
		var update = {};

		var to_null_out = {};
		var accounted_for = {};

		var i = levels.length;
		while (i--) {
			var o = levels[i];
			var n = updates[i];

			if (n) {
				for (var key in o) {
					if (!(key in n)) to_null_out[key] = 1;
				}

				for (var key in n) {
					if (!accounted_for[key]) {
						update[key] = n[key];
						accounted_for[key] = 1;
					}
				}

				levels[i] = n;
			} else {
				for (var key in o) {
					accounted_for[key] = 1;
				}
			}
		}

		for (var key in to_null_out) {
			if (!(key in update)) update[key] = undefined;
		}

		return update;
	}

	function blankObject() {
		return Object.create(null);
	}

	function destroy(detach) {
		this.destroy = noop;
		this.fire('destroy');
		this.set = noop;

		this._fragment.d(detach !== false);
		this._fragment = null;
		this._state = {};
	}

	function _differs(a, b) {
		return a != a ? b == b : a !== b || ((a && typeof a === 'object') || typeof a === 'function');
	}

	function fire(eventName, data) {
		var handlers =
			eventName in this._handlers && this._handlers[eventName].slice();
		if (!handlers) return;

		for (var i = 0; i < handlers.length; i += 1) {
			var handler = handlers[i];

			if (!handler.__calling) {
				try {
					handler.__calling = true;
					handler.call(this, data);
				} finally {
					handler.__calling = false;
				}
			}
		}
	}

	function flush(component) {
		component._lock = true;
		callAll(component._beforecreate);
		callAll(component._oncreate);
		callAll(component._aftercreate);
		component._lock = false;
	}

	function get() {
		return this._state;
	}

	function init(component, options) {
		component._handlers = blankObject();
		component._slots = blankObject();
		component._bind = options._bind;
		component._staged = {};

		component.options = options;
		component.root = options.root || component;
		component.store = options.store || component.root.store;

		if (!options.root) {
			component._beforecreate = [];
			component._oncreate = [];
			component._aftercreate = [];
		}
	}

	function on(eventName, handler) {
		var handlers = this._handlers[eventName] || (this._handlers[eventName] = []);
		handlers.push(handler);

		return {
			cancel: function() {
				var index = handlers.indexOf(handler);
				if (~index) handlers.splice(index, 1);
			}
		};
	}

	function set(newState) {
		this._set(assign({}, newState));
		if (this.root._lock) return;
		flush(this.root);
	}

	function _set(newState) {
		var oldState = this._state,
			changed = {},
			dirty = false;

		newState = assign(this._staged, newState);
		this._staged = {};

		for (var key in newState) {
			if (this._differs(newState[key], oldState[key])) changed[key] = dirty = true;
		}
		if (!dirty) return;

		this._state = assign(assign({}, oldState), newState);
		this._recompute(changed, this._state);
		if (this._bind) this._bind(changed, this._state);

		if (this._fragment) {
			this.fire("state", { changed: changed, current: this._state, previous: oldState });
			this._fragment.p(changed, this._state);
			this.fire("update", { changed: changed, current: this._state, previous: oldState });
		}
	}

	function _stage(newState) {
		assign(this._staged, newState);
	}

	function callAll(fns) {
		while (fns && fns.length) fns.shift()();
	}

	function _mount(target, anchor) {
		this._fragment[this._fragment.i ? 'i' : 'm'](target, anchor || null);
	}

	var proto = {
		destroy,
		get,
		fire,
		on,
		set,
		_recompute: noop,
		_set,
		_stage,
		_mount,
		_differs
	};

	/* src/VirtualList.html generated by Svelte v2.13.5 */

	function visible({ items, start, end }) {
		return items.slice(start, end).map((data, i) => {
			return { index: i + start, data };
		});
	}

	function data() {
		return { start: 0, end: 0, height: '100%', itemHeight: 0, _top: 0, _bottom: 0, _props: {} };
	}
	var methods = {
		initialise() {
			const { items, itemHeight, viewportHeight } = this.get();

			if (itemHeight) {
				this.heightMap = items.map(item => itemHeight);
				this.set({
					end: Math.min(items.length, Math.ceil(viewportHeight / itemHeight)),
					_bottom: items.length * itemHeight
				});
			} else {
				let height = 0;
				let i = 0;

				while (height < viewportHeight && i < items.length) {
					this.set({ end: i + 1 });

					const rowHeight = this.heightMap[i] = this.rows[i].offsetHeight;
					height += rowHeight;

					i += 1;
				}

				const end = i;
				const avg = Math.round(height / i);

				for (; i < items.length; i += 1) this.heightMap[i] = avg;

				this.set({
					_bottom: (items.length - end) * avg
				});
			}
		},

		refresh() {
			const { items, start, end, itemHeight, viewportHeight } = this.get();
			const { scrollTop } = this.refs.viewport;

			let paddingTop = 0;
			let offset = 0;
			let i = 0;

			if (!itemHeight) {
				for (let v = 0; v < this.rows.length; v += 1) {
					this.heightMap[start + v] = this.rows[v].offsetHeight;
				}
			}

			for (; i < items.length; i += 1) {
				if (!(i in this.heightMap)) break;

				offset += this.heightMap[i];
				if (offset > scrollTop) break;

				paddingTop = offset;
			}

			const newStart = i++;

			for (; i < items.length; i += 1) {
				if (offset >= scrollTop + viewportHeight) break;
				offset += this.heightMap[i];
			}

			const newEnd = i;

			if (newStart === start && newEnd === end) return;

			let paddingBottom = 0;
			for (; i < items.length; i += 1) paddingBottom += this.heightMap[i];

			this.set({
				_top: paddingTop,
				_bottom: paddingBottom,
				start: newStart,
				end: newEnd
			});

			if (newStart < start) {
				let d = 0;

				for (let i = newStart; i < start; i += 1) {
					const expectedHeight = this.heightMap[i];
					const actualHeight = this.rows[i - newStart].offsetHeight;

					d += actualHeight - expectedHeight;
				}

				this.refs.viewport.scrollTo(0, this.refs.viewport.scrollTop + d);
			}
		}
	};

	function oncreate() {
		const { items, _props } = this.get();
		const { container } = this.refs;

		const keys = Object.keys(this.options.data).filter(key => key !== 'items' && key !== 'component' && key !== 'height' && key !== 'itemHeight');
		if (keys.length) {
			const state = this.get();
			keys.forEach(key => {
				_props[key] = state[key];
			});
			this.set({ _props });
		}

		this.heightMap = [];
		this.rows = container.getElementsByClassName('row');

		if (items.length > 0) {
			this.initialise();
		}

		this.on('state', ({ changed, previous, current }) => {
			if (changed.items || changed.height || changed.itemHeight || changed.viewportHeight) {
				if (current.itemHeight && (changed.itemHeight || current.items.length > this.heightMap.length)) {
					this.heightMap = current.items.map(() => current.itemHeight);
				}

				else if (current.items.length > this.heightMap.length) {
					if (this.heightMap.length === 0) {
						this.initialise();
					} else {
						let height = 0;
						let i = 0;
						for (; i < this.heightMap.length; i += 1) height += this.heightMap[i];
						const avg = height / this.heightMap.length;
						for (; i < current.items.length; i += 1) this.heightMap[i] = avg;
					}
				}

				this.refresh();
			}

			if (keys.some(key => changed[key])) {
				const _props = {};
				keys.forEach(key => {
					_props[key] = current[key];
				});
				this.set({ _props });
			}
		});
	}
	function add_css() {
		var style = createElement("style");
		style.id = 'svelte-1xu8vv3-style';
		style.textContent = ".svelte-ref-viewport.svelte-1xu8vv3{position:relative;overflow-y:auto;-webkit-overflow-scrolling:touch}.row.svelte-1xu8vv3{overflow:hidden}";
		append(document.head, style);
	}

	function get_each_context(ctx, list, i) {
		const child_ctx = Object.create(ctx);
		child_ctx.item = list[i];
		child_ctx.each_value = list;
		child_ctx.item_index = i;
		return child_ctx;
	}

	function create_main_fragment(component, ctx) {
		var div1, div0, each_blocks_1 = [], each_lookup = blankObject(), div1_resize_listener;

		var each_value = ctx.visible;

		const get_key = ctx => ctx.item.index;

		for (var i = 0; i < each_value.length; i += 1) {
			let child_ctx = get_each_context(ctx, each_value, i);
			let key = get_key(child_ctx);
			each_blocks_1[i] = each_lookup[key] = create_each_block(component, key, child_ctx);
		}

		function div1_resize_handler() {
			component.set({ viewportHeight: div1.offsetHeight });
		}

		function scroll_handler(event) {
			component.refresh();
		}

		return {
			c() {
				div1 = createElement("div");
				div0 = createElement("div");

				for (i = 0; i < each_blocks_1.length; i += 1) each_blocks_1[i].c();
				setStyle(div0, "padding-top", "" + ctx._top + "px");
				setStyle(div0, "padding-bottom", "" + ctx._bottom + "px");
				component.root._beforecreate.push(div1_resize_handler);
				addListener(div1, "scroll", scroll_handler);
				setStyle(div1, "height", ctx.height);
				div1.className = "svelte-1xu8vv3 svelte-ref-viewport";
			},

			m(target, anchor) {
				insert(target, div1, anchor);
				append(div1, div0);

				for (i = 0; i < each_blocks_1.length; i += 1) each_blocks_1[i].m(div0, null);

				component.refs.container = div0;
				div1_resize_listener = addResizeListener(div1, div1_resize_handler);
				component.refs.viewport = div1;
			},

			p(changed, ctx) {
				const each_value = ctx.visible;
				each_blocks_1 = updateKeyedEach(each_blocks_1, component, changed, get_key, 1, ctx, each_value, each_lookup, div0, destroyBlock, create_each_block, "m", null, get_each_context);

				if (changed._top) {
					setStyle(div0, "padding-top", "" + ctx._top + "px");
				}

				if (changed._bottom) {
					setStyle(div0, "padding-bottom", "" + ctx._bottom + "px");
				}

				if (changed.height) {
					setStyle(div1, "height", ctx.height);
				}
			},

			d(detach) {
				if (detach) {
					detachNode(div1);
				}

				for (i = 0; i < each_blocks_1.length; i += 1) each_blocks_1[i].d();

				if (component.refs.container === div0) component.refs.container = null;
				div1_resize_listener.cancel();
				removeListener(div1, "scroll", scroll_handler);
				if (component.refs.viewport === div1) component.refs.viewport = null;
			}
		};
	}

	// (3:2) {#each visible as item (item.index)}
	function create_each_block(component, key_1, ctx) {
		var div, text;

		var switch_instance_spread_levels = [
			ctx._props,
			ctx.item.data
		];

		var switch_value = ctx.component;

		function switch_props(ctx) {
			var switch_instance_initial_data = {};
			for (var i = 0; i < switch_instance_spread_levels.length; i += 1) {
				switch_instance_initial_data = assign(switch_instance_initial_data, switch_instance_spread_levels[i]);
			}
			return {
				root: component.root,
				store: component.store,
				data: switch_instance_initial_data
			};
		}

		if (switch_value) {
			var switch_instance = new switch_value(switch_props(ctx));
		}

		return {
			key: key_1,

			first: null,

			c() {
				div = createElement("div");
				if (switch_instance) switch_instance._fragment.c();
				text = createText("\n\t\t\t");
				div.className = "row svelte-1xu8vv3";
				this.first = div;
			},

			m(target, anchor) {
				insert(target, div, anchor);

				if (switch_instance) {
					switch_instance._mount(div, null);
				}

				append(div, text);
			},

			p(changed, ctx) {
				var switch_instance_changes = (changed._props || changed.visible) ? getSpreadUpdate(switch_instance_spread_levels, [
					(changed._props) && ctx._props,
					(changed.visible) && ctx.item.data
				]) : {};

				if (switch_value !== (switch_value = ctx.component)) {
					if (switch_instance) {
						switch_instance.destroy();
					}

					if (switch_value) {
						switch_instance = new switch_value(switch_props(ctx));
						switch_instance._fragment.c();
						switch_instance._mount(div, null);
					} else {
						switch_instance = null;
					}
				}

				else if (switch_value) {
					switch_instance._set(switch_instance_changes);
				}
			},

			d(detach) {
				if (detach) {
					detachNode(div);
				}

				if (switch_instance) switch_instance.destroy();
			}
		};
	}

	function VirtualList(options) {
		init(this, options);
		this.refs = {};
		this._state = assign(data(), options.data);
		this._recompute({ items: 1, start: 1, end: 1 }, this._state);
		this._intro = true;

		if (!document.getElementById("svelte-1xu8vv3-style")) add_css();

		this._fragment = create_main_fragment(this, this._state);

		this.root._oncreate.push(() => {
			oncreate.call(this);
			this.fire("update", { changed: assignTrue({}, this._state), current: this._state });
		});

		if (options.target) {
			this._fragment.c();
			this._mount(options.target, options.anchor);

			flush(this);
		}
	}

	assign(VirtualList.prototype, proto);
	assign(VirtualList.prototype, methods);

	VirtualList.prototype._recompute = function _recompute(changed, state) {
		if (changed.items || changed.start || changed.end) {
			if (this._differs(state.visible, (state.visible = visible(state)))) changed.visible = true;
		}
	};

	return VirtualList;

})));
